@@ -1,18 +1,29 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE FlexibleContexts, FlexibleInstances #-}
 {-# LANGUAGE TypeOperators #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE LambdaCase, MultiWayIf #-}
--- | Vty supports a configuration file format and associated 'Config'
--- data type. The 'Config' can be provided to 'mkVty' to customize the
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FlexibleContexts #-}
+-- | Vty supports a configuration file format and provides a
+-- corresponding 'VtyUserConfig' data type. The 'VtyUserConfig' can be
+-- provided to platform packages' @mkVty@ functions to customize the
 -- application's use of Vty.
 --
--- Lines in config files that fail to parse are ignored. Later entries
--- take precedence over earlier ones.
---
 -- = Debug
 --
+-- == @colorMode@
+--
+-- Format:
+--
+-- @
+--  colorMode \"<ColorMode8|ColorMode16|ColorMode240 <int>|FullColor>\"
+-- @
+--
+-- The preferred color mode to use, chosen from the constructors of the
+-- 'ColorMode' type. If absent, the backend driver may detect and choose
+-- an appropriate color mode. Implementor's note: backend packages
+-- should respect this setting when it is present even when their
+-- detection indicates that a different color mode should be used.
+--
 -- == @debugLog@
 --
 -- Format:
@@ -73,27 +84,23 @@
 --   widthMap \"xterm\" \"\/home\/user\/.vty\/xterm\_map.dat\"
 -- @
 --
--- This directive specifies the path to a Unicode character width
--- map (the second argument) that should be loaded and used when
--- the value of TERM matches the first argument. Unicode character
--- width maps can be produced either by running the provided binary
--- @vty-build-width-table@ or by calling the library routine
--- 'Graphics.Vty.UnicodeWidthTable.Query.buildUnicodeWidthTable'. The
--- 'Graphics.Vty.mkVty' function will use these configuration settings
--- to attempt to load and install the specified width map. See the
--- documentation for 'Graphics.Vty.mkVty' for details.
+-- This directive specifies the path to a Unicode character
+-- width map (the second argument) that should correspond to
+-- the terminal named by first argument. Unicode character
+-- width maps can be produced either by running platform
+-- packages' width table tools or by calling the library routine
+-- 'Graphics.Vty.UnicodeWidthTable.Query.buildUnicodeWidthTable'. Vty
+-- platform packages should use these configuration settings to attempt
+-- to load and install the specified width map.
 module Graphics.Vty.Config
   ( InputMap
-  , Config(..)
-  , VtyConfigurationError(..)
+  , VtyUserConfig(..)
   , userConfig
   , overrideEnvConfig
-  , standardIOConfig
+  , currentTerminalName
   , runParseConfig
   , parseConfigFile
   , defaultConfig
-  , getTtyEraseChar
-  , currentTerminalName
 
   , vtyConfigPath
   , widthTableFilename
@@ -110,7 +117,7 @@ import Prelude
 
 import Control.Applicative hiding (many)
 
-import Control.Exception (catch, IOException, Exception(..), throwIO)
+import Control.Exception (catch, IOException)
 import Control.Monad (liftM, guard, void)
 
 import qualified Data.ByteString as BS
@@ -120,10 +127,10 @@ import Data.Monoid (Monoid(..))
 #if !(MIN_VERSION_base(4,11,0))
 import Data.Semigroup (Semigroup(..))
 #endif
-import Data.Typeable (Typeable)
+import Text.Read (readMaybe)
 
+import Graphics.Vty.Attributes.Color (ColorMode(..))
 import Graphics.Vty.Input.Events
-import Graphics.Vty.Attributes.Color (ColorMode(..), detectColorMode)
 
 import GHC.Generics
 
@@ -132,111 +139,73 @@ import System.Directory ( getAppUserDataDirectory, doesFileExist
                         )
 import System.Environment (lookupEnv)
 import System.FilePath ((</>), takeDirectory)
-import System.Posix.IO (stdInput, stdOutput)
-import System.Posix.Types (Fd(..))
-import Foreign.C.Types (CInt(..), CChar(..))
 
 import Text.Parsec hiding ((<|>))
 import Text.Parsec.Token ( GenLanguageDef(..) )
 import qualified Text.Parsec.Token as P
 
--- | Type of errors that can be thrown when configuring VTY
-data VtyConfigurationError =
-    VtyMissingTermEnvVar
-    -- ^ TERM environment variable not set
-    deriving (Show, Eq, Typeable)
-
-instance Exception VtyConfigurationError where
-    displayException VtyMissingTermEnvVar = "TERM environment variable not set"
-
 -- | Mappings from input bytes to event in the order specified. Later
 -- entries take precedence over earlier in the case multiple entries
 -- have the same byte string.
 type InputMap = [(Maybe String, String, Event)]
 
--- | A Vty configuration.
-data Config =
-    Config { vmin  :: Maybe Int
-           -- ^ The default is 1 character.
-           , vtime :: Maybe Int
-           -- ^ The default is 100 milliseconds, 0.1 seconds.
-           , mouseMode :: Maybe Bool
-           -- ^ The default is False.
-           , bracketedPasteMode :: Maybe Bool
-           -- ^ The default is False.
-           , debugLog :: Maybe FilePath
-           -- ^ Debug information is appended to this file if not
-           -- Nothing.
-           , inputMap :: InputMap
-           -- ^ The (input byte, output event) pairs extend the internal
-           -- input table of VTY and the table from terminfo.
-           --
-           -- See "Graphics.Vty.Config" module documentation for
-           -- documentation of the @map@ directive.
-           , inputFd :: Maybe Fd
-           -- ^ The input file descriptor to use. The default is
-           -- 'System.Posix.IO.stdInput'
-           , outputFd :: Maybe Fd
-           -- ^ The output file descriptor to use. The default is
-           -- 'System.Posix.IO.stdOutput'
-           , termName :: Maybe String
-           -- ^ The terminal name used to look up terminfo capabilities.
-           -- The default is the value of the TERM environment variable.
-           , termWidthMaps :: [(String, FilePath)]
-           -- ^ Terminal width map files.
-           , allowCustomUnicodeWidthTables :: Maybe Bool
-           -- ^ Whether to permit custom Unicode width table loading by
-           -- 'Graphics.Vty.mkVty'. @'Just' 'False'@ indicates that
-           -- table loading should not be performed. Other values permit
-           -- table loading.
-           --
-           -- If a table load is attempted and fails, information
-           -- about the failure will be logged to the debug log if the
-           -- configuration specifies one. If no custom table is loaded
-           -- (or if a load fails), the built-in character width table
-           -- will be used.
-           , colorMode :: Maybe ColorMode
-           -- ^ The color mode used to know how many colors the terminal
-           -- supports.
-           }
-           deriving (Show, Eq)
+-- | A Vty core library configuration. Platform-specific details are not
+-- included in the VtyUserConfig.
+data VtyUserConfig =
+    VtyUserConfig { configDebugLog :: Maybe FilePath
+                  -- ^ Debug information is appended to this file if not
+                  -- Nothing.
+                  , configInputMap :: InputMap
+                  -- ^ The (input byte, output event) pairs extend the internal
+                  -- input table of VTY and the table from terminfo.
+                  --
+                  -- See "Graphics.Vty.Config" module documentation for
+                  -- documentation of the @map@ directive.
+                  , configTermWidthMaps :: [(String, FilePath)]
+                  -- ^ Terminal width map files.
+                  , configAllowCustomUnicodeWidthTables :: Maybe Bool
+                  -- ^ Whether to permit custom Unicode width table loading by
+                  -- 'Graphics.Vty.mkVty'. @'Just' 'False'@ indicates that
+                  -- table loading should not be performed. Other values permit
+                  -- table loading.
+                  --
+                  -- If a table load is attempted and fails, information
+                  -- about the failure will be logged to the debug log if the
+                  -- configuration specifies one. If no custom table is loaded
+                  -- (or if a load fails), the built-in character width table
+                  -- will be used.
+                  , configPreferredColorMode :: Maybe ColorMode
+                  -- ^ Preferred color mode. If set, this should
+                  -- override platform color mode detection.
+                  }
+                  deriving (Show, Eq)
 
-defaultConfig :: Config
+defaultConfig :: VtyUserConfig
 defaultConfig = mempty
 
-instance Semigroup Config where
+instance Semigroup VtyUserConfig where
     c0 <> c1 =
         -- latter config takes priority for everything but inputMap
-        Config { vmin = vmin c1 <|> vmin c0
-               , vtime = vtime c1 <|> vtime c0
-               , mouseMode = mouseMode c1
-               , bracketedPasteMode = bracketedPasteMode c1
-               , debugLog = debugLog c1 <|> debugLog c0
-               , inputMap = inputMap c0 <> inputMap c1
-               , inputFd = inputFd c1 <|> inputFd c0
-               , outputFd = outputFd c1 <|> outputFd c0
-               , termName = termName c1 <|> termName c0
-               , termWidthMaps = termWidthMaps c1 <|> termWidthMaps c0
-               , allowCustomUnicodeWidthTables =
-                   allowCustomUnicodeWidthTables c1 <|> allowCustomUnicodeWidthTables c0
-               , colorMode = colorMode c1 <|> colorMode c0
-               }
+        VtyUserConfig { configDebugLog =
+                          configDebugLog c1 <|> configDebugLog c0
+                      , configInputMap =
+                          configInputMap c0 <> configInputMap c1
+                      , configTermWidthMaps =
+                          configTermWidthMaps c1 <|> configTermWidthMaps c0
+                      , configAllowCustomUnicodeWidthTables =
+                          configAllowCustomUnicodeWidthTables c1 <|> configAllowCustomUnicodeWidthTables c0
+                      , configPreferredColorMode =
+                          configPreferredColorMode c1 <|> configPreferredColorMode c0
+                      }
 
-instance Monoid Config where
+instance Monoid VtyUserConfig where
     mempty =
-        Config { vmin = Nothing
-               , vtime = Nothing
-               , mouseMode = Nothing
-               , bracketedPasteMode = Nothing
-               , debugLog = mempty
-               , inputMap = mempty
-               , inputFd = Nothing
-               , outputFd = Nothing
-               , termName = Nothing
-               , termWidthMaps = []
-               , allowCustomUnicodeWidthTables = Nothing
-               , colorMode = Nothing
-               }
+        VtyUserConfig { configDebugLog = mempty
+                      , configInputMap = mempty
+                      , configTermWidthMaps = []
+                      , configAllowCustomUnicodeWidthTables = Nothing
+                      , configPreferredColorMode = Nothing
+                      }
 #if !(MIN_VERSION_base(4,11,0))
     mappend = (<>)
 #endif
@@ -253,7 +222,8 @@ vtyConfigFileEnvName :: String
 vtyConfigFileEnvName = "VTY_CONFIG_FILE"
 
 -- | Load a configuration from 'vtyConfigPath' and @$VTY_CONFIG_FILE@.
-userConfig :: IO Config
+-- If none is found, build a default configuration.
+userConfig :: IO VtyUserConfig
 userConfig = do
     configFile <- vtyConfigPath >>= parseConfigFile
     overrideConfig <- maybe (return defaultConfig) parseConfigFile =<<
@@ -279,37 +249,21 @@ terminalWidthTablePath = do
         Just term -> do
             return $ Just $ dataDir </> widthTableFilename term
 
-overrideEnvConfig :: IO Config
+overrideEnvConfig :: IO VtyUserConfig
 overrideEnvConfig = do
     d <- lookupEnv "VTY_DEBUG_LOG"
-    return $ defaultConfig { debugLog = d }
+    return $ defaultConfig { configDebugLog = d }
 
--- | Configures VTY using defaults suitable for terminals. This function
--- can raise 'VtyConfigurationError'.
-standardIOConfig :: IO Config
-standardIOConfig = do
-    mb <- lookupEnv termVariable
-    case mb of
-      Nothing -> throwIO VtyMissingTermEnvVar
-      Just t -> do
-        mcolorMode <- detectColorMode t
-        return defaultConfig
-          { vmin               = Just 1
-          , mouseMode          = Just False
-          , bracketedPasteMode = Just False
-          , vtime              = Just 100
-          , inputFd            = Just stdInput
-          , outputFd           = Just stdOutput
-          , termName           = Just t
-          , colorMode          = Just mcolorMode
-          }
-
-parseConfigFile :: FilePath -> IO Config
+-- | Parse a Vty configuration file.
+--
+-- Lines in config files that fail to parse are ignored. Later entries
+-- take precedence over earlier ones.
+parseConfigFile :: FilePath -> IO VtyUserConfig
 parseConfigFile path = do
     catch (runParseConfig path <$> BS.readFile path)
           (\(_ :: IOException) -> return defaultConfig)
 
-runParseConfig :: String -> BS.ByteString -> Config
+runParseConfig :: String -> BS.ByteString -> VtyUserConfig
 runParseConfig name cfgTxt =
   case runParser parseConfig () name cfgTxt of
     Right cfg -> cfg
@@ -337,7 +291,7 @@ configLanguage = LanguageDef
 configLexer :: Monad m => P.GenTokenParser BS.ByteString () m
 configLexer = P.makeTokenParser configLanguage
 
-mapDecl :: Parser Config
+mapDecl :: Parser VtyUserConfig
 mapDecl = do
     "map" <- P.identifier configLexer
     termIdent <- (char '_' >> P.whiteSpace configLexer >> return Nothing)
@@ -345,28 +299,34 @@ mapDecl = do
     bytes     <- P.stringLiteral configLexer
     key       <- parseValue
     modifiers <- parseValue
-    return defaultConfig { inputMap = [(termIdent, bytes, EvKey key modifiers)] }
+    return defaultConfig { configInputMap = [(termIdent, bytes, EvKey key modifiers)] }
 
-debugLogDecl :: Parser Config
+debugLogDecl :: Parser VtyUserConfig
 debugLogDecl = do
     "debugLog" <- P.identifier configLexer
     path       <- P.stringLiteral configLexer
-    return defaultConfig { debugLog = Just path }
+    return defaultConfig { configDebugLog = Just path }
 
-widthMapDecl :: Parser Config
+colorModeDecl :: Parser VtyUserConfig
+colorModeDecl = do
+    "colorMode" <- P.identifier configLexer
+    mode <- P.stringLiteral configLexer
+    return defaultConfig { configPreferredColorMode = readMaybe mode }
+
+widthMapDecl :: Parser VtyUserConfig
 widthMapDecl = do
     "widthMap" <- P.identifier configLexer
     tName <- P.stringLiteral configLexer
     path <- P.stringLiteral configLexer
-    return defaultConfig { termWidthMaps = [(tName, path)] }
+    return defaultConfig { configTermWidthMaps = [(tName, path)] }
 
 ignoreLine :: Parser ()
 ignoreLine = void $ manyTill anyChar newline
 
-parseConfig :: Parser Config
+parseConfig :: Parser VtyUserConfig
 parseConfig = liftM mconcat $ many $ do
     P.whiteSpace configLexer
-    let directives = [try mapDecl, try debugLogDecl, try widthMapDecl]
+    let directives = [try mapDecl, try debugLogDecl, try widthMapDecl, try colorModeDecl]
     choice directives <|> (ignoreLine >> return defaultConfig)
 
 class    Parse a        where parseValue :: Parser a
@@ -413,32 +373,22 @@ instance (GParseAlts f, GParseAlts g) => GParseAlts (f :+: g) where
 
 instance GParseAlts V1 where gparseAlts _ = fail "GParse: V1"
 
-foreign import ccall "vty_get_tty_erase" cGetTtyErase :: Fd -> IO CChar
-
--- | Get the "erase" character for the terminal attached to the
--- specified file descriptor. This is the character configured by 'stty
--- erase'. If the call to 'tcgetattr' fails, this will return 'Nothing'.
--- Otherwise it will return the character that has been configured to
--- indicate the canonical mode ERASE behavior. That character can then
--- be added to the table of strings that we interpret to mean Backspace.
---
--- For more details, see:
---
--- * https://www.gnu.org/software/libc/manual/html_node/Canonical-or-Not.html
--- * https://www.gsp.com/cgi-bin/man.cgi?section=1&topic=stty
--- * https://github.com/matterhorn-chat/matterhorn/issues/565
-getTtyEraseChar :: Fd -> IO (Maybe Char)
-getTtyEraseChar fd = do
-    c <- cGetTtyErase fd
-    if c /= 0
-       then return $ Just $ toEnum $ fromEnum c
-       else return Nothing
-
+-- | The result of a configuration change attempt made by
+-- 'addConfigWidthMap'.
 data ConfigUpdateResult =
     ConfigurationCreated
+    -- ^ A new configuration file file was written with the new width
+    -- table entry.
     | ConfigurationModified
+    -- ^ An existing configuration file was modified with the new width
+    -- table entry.
     | ConfigurationConflict String
+    -- ^ The attempted width table entry could not be written to the
+    -- configuration due to a conflict; the argument here is the width
+    -- table file path for the conflicting entry.
     | ConfigurationRedundant
+    -- ^ No change was made because the existing configuration already
+    -- contains the specified mapping.
     deriving (Eq, Show)
 
 -- | Add a @widthMap@ directive to the Vty configuration file at the
@@ -452,8 +402,8 @@ data ConfigUpdateResult =
 -- If the configuration path does not exist, a new configuration file
 -- will be created and any directories in the path will also be created.
 --
--- This returns @True@ if the configuration was created or modified and
--- @False@ otherwise. This does not handle exceptions raised by file or
+-- This returns a 'ConfigUpdateResult' indicating the change to the
+-- configuration. This does not handle exceptions raised by file or
 -- directory permissions issues.
 addConfigWidthMap :: FilePath
                   -- ^ The configuration file path of the configuration
@@ -479,9 +429,9 @@ addConfigWidthMap configPath term tablePath = do
 
         updateConfig = do
             config <- parseConfigFile configPath
-            if (term, tablePath) `elem` termWidthMaps config
+            if (term, tablePath) `elem` configTermWidthMaps config
                then return ConfigurationRedundant
-               else case lookup term (termWidthMaps config) of
+               else case lookup term (configTermWidthMaps config) of
                    Just other -> return $ ConfigurationConflict other
                    Nothing -> do
                        appendFile configPath directive
