@@ -3,10 +3,12 @@
 {-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE NamedFieldPuns #-}
 {-# LANGUAGE RankNTypes #-}
-{-# LANGUAGE TemplateHaskell #-}
 
 -- | Transforms an image into rows of operations.
-module Graphics.Vty.PictureToSpans where
+module Graphics.Vty.PictureToSpans
+  ( displayOpsForPic
+  )
+where
 
 import Graphics.Vty.Attributes (Attr, currentAttr)
 import Graphics.Vty.Image
@@ -16,7 +18,7 @@ import Graphics.Vty.Span
 
 import Lens.Micro
 import Lens.Micro.Mtl
-import Lens.Micro.TH
+import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State.Strict hiding ( state )
 import Control.Monad.ST.Strict
@@ -29,8 +31,6 @@ import qualified Data.Text.Lazy as TL
 
 type MRowOps s = MVector s SpanOps
 
-type MSpanOps s = MVector s SpanOp
-
 -- transform plus clip. More or less.
 data BlitState = BlitState
     -- we always snoc to the operation vectors. Thus the columnOffset =
@@ -53,14 +53,34 @@ data BlitState = BlitState
     , _remainingRows :: Int
     }
 
-makeLenses ''BlitState
+columnOffset :: Lens' BlitState Int
+columnOffset = lens _columnOffset (\e v -> e { _columnOffset = v })
+
+rowOffset :: Lens' BlitState Int
+rowOffset = lens _rowOffset (\e v -> e { _rowOffset = v })
+
+skipColumns :: Lens' BlitState Int
+skipColumns = lens _skipColumns (\e v -> e { _skipColumns = v })
+
+skipRows :: Lens' BlitState Int
+skipRows = lens _skipRows (\e v -> e { _skipRows = v })
+
+remainingColumns :: Lens' BlitState Int
+remainingColumns = lens _remainingColumns (\e v -> e { _remainingColumns = v })
+
+remainingRows :: Lens' BlitState Int
+remainingRows = lens _remainingRows (\e v -> e { _remainingRows = v })
 
 data BlitEnv s = BlitEnv
     { _region :: DisplayRegion
     , _mrowOps :: MRowOps s
     }
 
-makeLenses ''BlitEnv
+region :: Lens' (BlitEnv s) DisplayRegion
+region = lens _region (\e r -> e { _region = r })
+
+mrowOps :: Lens' (BlitEnv s) (MRowOps s)
+mrowOps = lens _mrowOps (\e r -> e { _mrowOps = r })
 
 type BlitM s a = ReaderT (BlitEnv s) (StateT BlitState (ST s)) a
 
@@ -69,13 +89,6 @@ type BlitM s a = ReaderT (BlitEnv s) (StateT BlitState (ST s)) a
 displayOpsForPic :: Picture -> DisplayRegion -> DisplayOps
 displayOpsForPic pic r = Vector.create (combinedOpsForLayers pic r)
 
--- | Returns the DisplayOps for an image rendered to a window the size
--- of the image.
---
--- largerly used only for debugging.
-displayOpsForImage :: Image -> DisplayOps
-displayOpsForImage i = displayOpsForPic (picForImage i) (imageWidth i, imageHeight i)
-
 -- | Produces the span ops for each layer then combines them.
 combinedOpsForLayers :: Picture -> DisplayRegion -> ST s (MRowOps s)
 combinedOpsForLayers pic r
@@ -83,7 +96,7 @@ combinedOpsForLayers pic r
     | otherwise = do
         layerOps <- mapM (`buildSpans` r) (picLayers pic)
         case layerOps of
-            []    -> fail "empty picture"
+            []    -> error "empty picture"
             [ops] -> substituteSkips (picBackground pic) ops
             -- instead of merging ops after generation the merging can
             -- be performed as part of snocOp.
@@ -113,7 +126,7 @@ substituteSkips (Background {backgroundChar, backgroundAttr}) ops = do
     -- At this point we decide if the background character is single
     -- column or not. obviously, single column is easier.
     case safeWcwidth backgroundChar of
-        w | w == 0 -> fail $ "invalid background character " ++ show backgroundChar
+        w | w == 0 -> error $ "invalid background character " ++ show backgroundChar
           | w == 1 -> do
                 forM_ [0 .. MVector.length ops - 1] $ \row -> do
                     rowOps <- MVector.read ops row
@@ -220,18 +233,18 @@ startImageBuild image = do
 
 isOutOfBounds :: Image -> BlitState -> Bool
 isOutOfBounds i s
-    | s ^. remainingColumns <= 0              = True
-    | s ^. remainingRows    <= 0              = True
+    | s ^. remainingColumns <= 0             = True
+    | s ^. remainingRows    <= 0             = True
     | s ^. skipColumns      >= imageWidth i  = True
     | s ^. skipRows         >= imageHeight i = True
     | otherwise = False
 
 -- | This adds an image that might be partially clipped to the output
 -- ops.
---
 -- This is a very touchy algorithm. Too touchy. For instance, the
--- CropRight and CropBottom implementations are odd. They pass the
--- current tests but something seems terribly wrong about all this.
+-- Crop implementations is odd. They pass the current tests but
+-- something seems terribly wrong about all this.
+--
 addMaybeClipped :: forall s . Image -> BlitM s ()
 addMaybeClipped EmptyImage = return ()
 addMaybeClipped (HorizText a textStr ow _cw) = do
@@ -266,23 +279,13 @@ addMaybeClipped BGFill {outputWidth, outputHeight} = do
         outputHeight' = min (outputHeight - s^.skipRows   ) (s^.remainingRows)
     y <- use rowOffset
     forM_ [y..y+outputHeight'-1] $ snocOp (Skip outputWidth')
-addMaybeClipped CropRight {croppedImage, outputWidth} = do
-    s <- use skipColumns
-    r <- use remainingColumns
-    let x = outputWidth - s
-    when (x < r) $ remainingColumns .= x
-    addMaybeClipped croppedImage
-addMaybeClipped CropLeft {croppedImage, leftSkip} = do
+addMaybeClipped Crop {croppedImage, leftSkip, topSkip, outputWidth, outputHeight} = do
+    sx <- use skipColumns
     skipColumns += leftSkip
-    addMaybeClipped croppedImage
-addMaybeClipped CropBottom {croppedImage, outputHeight} = do
-    s <- use skipRows
-    r <- use remainingRows
-    let x = outputHeight - s
-    when (x < r) $ remainingRows .= x
-    addMaybeClipped croppedImage
-addMaybeClipped CropTop {croppedImage, topSkip} = do
+    modifying remainingColumns (min (outputWidth - sx))
+    sy <- use skipRows
     skipRows += topSkip
+    modifying remainingRows (min (outputHeight - sy))
     addMaybeClipped croppedImage
 
 addMaybeClippedJoin :: forall s . String
@@ -296,7 +299,7 @@ addMaybeClippedJoin :: forall s . String
                        -> BlitM s ()
 addMaybeClippedJoin name skip remaining offset i0Dim i0 i1 size = do
     state <- get
-    when (state^.remaining <= 0) $ fail $ name ++ " with remaining <= 0"
+    when (state^.remaining <= 0) $ error $ name ++ " with remaining <= 0"
     case state^.skip of
         s | s > size -> put $ state & skip %~ subtract size
           | s == 0    -> if state^.remaining > i0Dim
@@ -316,9 +319,9 @@ addMaybeClippedJoin name skip remaining offset i0Dim i0 i1 size = do
           | s >= i0Dim -> do
                 put $ state & skip %~ subtract i0Dim
                 addMaybeClipped i1
-        _ -> fail $ name ++ " has unhandled skip class"
+        _ -> error $ name ++ " has unhandled skip class"
 
-addUnclippedText :: Attr -> DisplayText -> BlitM s ()
+addUnclippedText :: Attr -> TL.Text -> BlitM s ()
 addUnclippedText a txt = do
     let op = TextSpan a usedDisplayColumns
                       (fromIntegral $ TL.length txt)
@@ -344,5 +347,5 @@ snocOp !op !row = do
         ops <- MVector.read theMrowOps row
         let ops' = Vector.snoc ops op
         when (spanOpsAffectedColumns ops' > regionWidth theRegion)
-             $ fail $ "row " ++ show row ++ " now exceeds region width"
+             $ error $ "row " ++ show row ++ " now exceeds region width"
         MVector.write theMrowOps row ops'
